// teethtime.c

const char rev=0x10;  // rev 1.0 (01-Jan-05)

#include <iot2313v.h>
#pragma interrupt_handler timer:iv_TIMER1_COMPA
#pragma interrupt_handler button:iv_INT0
#pragma global_register led:20 limit:21 j:22 k:23

typedef unsigned char uchar;
typedef unsigned char bool;
typedef unsigned int  uint;
#define BIT(x) (1<<(x))
#define InterruptEnable 0x80
#define clk1024 5        // prescaler mode = clk/1024
#define SpeakerOFF 0x40  // Speaker on PD6
#define Pullup 0x04      // for pushbutton (PD2)
#define NPAR 8  // number of user parameters
                   // 0: limit (high nibble)
                   // 1: limit (low nibble)
                   // 2: period (high nibble)
                   // 3: period (low nibble)
                   // 4: bcd
                   // 5: pitch
                   // 6: tempo
                   // 7: song
#define NEXIT 4 // number of exit codes
                   // 8:  exit and save(0)
                   // 9:  exit and save(1)
                   // 10: exit and recall(0)
                   // 11: exit and recall(1)
#define SetupCode 1 // To go to setup, press button after the 1st count

// Global variables ---------------------------------
uchar go;
uchar isong;
uchar pressed;
uchar beepon;
uchar led;
uchar j,k;
                 // User parameters ---------------
uchar limit;     // max count
uchar period=16; // increment every second (units = 1/16 sec)
uchar bcd;       // 0/1 Binary up/down.  2/3 BCD up/down. Default = 0 = binary/up
uchar pitch=3;   // default pitch (smaller numbers for higher pitch)
uchar tempo=8;   // default tempo (smaller numbers for faster temo)
uchar song;      // default ring tone (0 does a 16 way rotate)

//                      0    1    2    3    4    5    6    7    8    9    10   11   12   13   14   15  16  17  18  19  20  21  22  23  24
// #/- = sharp/flat     A-   A    B-   B    C    C#   D    E-   E    F    F#   G    G#   a    a#   b   c   c#  d   d#  e   f   f#  g   g# 
const uchar scale[] = {240, 226, 216, 200, 192, 180, 169, 160, 150, 144, 135, 128, 120, 113, 108, 100, 96, 90, 84, 80, 75, 72, 67, 64, 60 };
                   // rest identifier:  h     i     j    k    l   m n o
                   // duration:        1/32  1/16  1/8  1/4  1/2  1 2 4
const char song0[] = "2E4-e4-db1-c1de-E4-c4b8-jC4a4-GE1-F1Ga-FD1E1-FGE4-"; // Over the Rainbow
const char song1[] = "2AE1iE1iEjE1iE1iEjEjF3E1D4jDjDjEDCD3C1BA#BE6";       // Chim Chim Cher-ee
const char song2[] = "2D4jDajajaE3F1ED6nacd4cabGa4";       // Scarborough Fair
const char song3[] = "1edb4d2G3jGED3GjGjGjGjbjbja2G6";     // Swing Low
const char song4[] = "2GEG3jGEG3jGEG3a1GEG4jEFGF4DGF4DC6"; // The Caissons go Rolling
const char song5[] = "1CEFG3jCEFG3jCEFG2E2C2E2D6";         // When the Saints
const char song6[] = "2CjCEGc6a8maFGaG9mCCEGjG6D9mEFEDC8"; // On Top of Old Smoky
const char song7[] = "2DGb-a1G1F#aG6D6jDGb-a1G1F#aG6";     // Sunrise, Sunset
const char song8[] = "1GF#GaGmambabcbmcmd2GcbmamG6";       // Jacques Saint-Luc: The Lute Player
const char song9[] = "1d2c2#d2e2c2b2c2d2bdgf#edcba2oD6E2C6D2BmDmEmF#mG2";       // Haslinger Sonatina
const char songA[] = "1g4fedcg4fedcEFGmFmEmD2GmEmCmg4fedcg4fedcEFGmFmEmD2GmC4"; // Bach Musette in D major
const char songB[] = "4dkekckCkG";   // Close Encounters 3rd kind
const char songC[] = "1CDEFGabc";    // Major scale
const char songD[] = "2CEGc";        // Major arpeggio
const char songE[] = "1ala";         // Two beeps
const char songF[] = "1c";           // One beep
const char *songN[] = {song0, song1, song2, song3, song4, song5, song6, song7, song8, song9, songA, songB, songC, songD, songE, songF};
//                      A  B  C  D  E  F  G   a   b   c   d   e   f   g
const char sc[]     = { 1, 3, 4, 6, 8, 9, 11, 13, 15, 16, 18, 20, 21, 23};
// ---------------------------------------------------

void beep(uchar freq, uint duration);
void delayMS(uchar msec);
uint mulj(uint a);
uchar EEPROMread(char adr);
int EEPROMwrite(char adr, uchar byte);

void main()
{ const char *p; // points to currently played note of ring tone
  uchar a;       // default note duration
  limit = 120;   // default limit = 120 seconds
  // Set up hardware ------------------------------------
  DDRB = 0xFF;  // Port B drive the LEDs (outputs)
  DDRD = 0xFB;  // Set PORTD to all outputs except PD2 (button)
  PORTD = SpeakerOFF + Pullup; // turn speaker off; pullup for button
  beepon = SpeakerOFF;      // enable speaker (despite appearances:)
  MCUCR = BIT(SE)+BIT(SM0); // Sleep Enable, Sleep Mode = power down
  TIMSK = BIT(OCIE1A);      // Output Compare Interrupt Enable A
  GIMSK = BIT(INT0);        // Enable INT0 interrupt
  SREG |= InterruptEnable;  // Global interrupt enable
  while(1) // Main Loop ------------- Forever -----------------------------
  { go = 0;                        // disable counting
    j=250; OCR1A = mulj(period);   // Compare A = 4096MHz/(1024*250*period)
    led=pressed=TCNT1H=TCNT1L=0;   // zero led count, button flag, Timer1
    TCCR1B = BIT(WGM12)+clk1024;   // Start Timer1, WGM12=Clear Timer on Compare
    do { a = j = led;
         if (bcd&1)                           // here for down counting
         { j = limit - j;                     // display time remaining
           if ((bcd>1) && ((j&15)>9)) j-= 6;  // bcd correction
         }
         PORTB = ~j;               // drive leds
       }
       while (a!=limit);           // wait for limit while updating LEDs
    if (!pressed)                  // Play the song when the limit is reached
    { led = song;                  // get song parameter
      if (!led) { led = isong; isong = 15 & (isong+1); }  // here for song rotate mode
      p = songN[led];              // point to selected song
      a = *p++ - '0';              // first character is default duration
      while (k=*p++)               // process the rest of the song
      { if (pressed) break;        // the button aborts the song
        k-='A'; if (k>6) k-=25;    // lower case a-g are an octave higher than A-G
        if ((j=k-sizeof(sc))<8) { beepon=0; beep(225,2<<j); beepon=SpeakerOFF; }  // rests
        else  // if not a rest, play the note ---
        { k = sc[k]; // get index to scale[] without taking account of sharps/flats
          led = *p;  j = a;                          // j will be the note duration
          if ((led>='1') && (led<='9'))              // duration code first
          { j = led-'0'; led = *++p; }
          if      (led=='#') { k++; goto incp; }     // then accidental codes
          else if (led=='-') { k--; incp: ++p; }
          led=scale[24-k]; beep(scale[k],mulj(led)); // play the note
        }
      }
    }
    led = 0; PORTB = 0xFF; // Turn LEDs off
    asm("sleep");          // Go to sleep
  }
}

void beep(uchar freq, uint duration) // play a note (or a rest)
{ uchar c = SpeakerOFF + Pullup;
  uchar a,b;
  j = tempo; duration = 0xFFFE & (mulj(duration)>>2);
  while (duration--)  // note: duration must be even
    { PORTD = (c-=beepon); for(j=freq; j; j--) for(a=pitch; a; a--); }
}

uint mulj(uint a)  // unsigned multiply: 8 bit j by 16 bit a
{ uint ans = 0;
  while (j) { if (j&1) ans+=a; a<<=1; j>>=1; }
  return(ans);
}

void delayMS(uchar msec)  // delay msec milliseconds
{ uint a;
  while (msec--) for (a=453; a; a--);
}

void timer()     // Timer B interrupt
{ led+=go; go=1;  if ((bcd>1) && ((led&15)==10)) led += 6; }

bool bdown() { return !(PIND & Pullup); }  // returns true if button is down

void button()           // INT2 interrupt (on button press)
{ uchar e[NPAR+NEXIT];  // parameters start at e[1]
  TCCR1B = 0;           // Stop Timer1
  delayMS(10);          // debounce
  while (bdown());      // Wait for button release
  beep(led?80:120,10);  // Beep
  delayMS(10);          // Debounce
  if (led==SetupCode)
  { delayMS(200); beep(120,10);
    e[0]  = limit >> 4;
    e[1]  = limit & 15;
    e[2]  = period >> 4;
    e[3]  = period & 15;
    e[4]  = bcd;
    e[5]  = pitch;
    e[6]  = tempo;
    e[7]  = song;
    e[8]  = 0;   // save(0) and exit setup
    e[9]  = 0;   // save(1) and exit setup
    e[10] = 0;   // recall(0) and exit setup
    e[11] = 0;   // recall(1) and exit setup
    k = 0; // k identifies which parameter (e[0] thru e[NPAR-1])
           // is being displayed or modified or which exit code
           // e[NPAR] thru e[NPAR+NEXIT-1] will be used.
    while(1)     // do until the button is pushed while k is zero
    { while(1)   // cycle thru all parameters until button is pressed
      { PORTB = ~(k  + (e[k]<<4));  // low/high nibble is param number/value
        for (led=240; led; led--)   // cycle at 1.2 sec per parameter (250*.005)
        { delayMS(5); if (bdown()) goto ma; }
        if (++k == sizeof(e)) k=0;  // wrap from last exit code back to zero
      }
      ma: delayMS(10);           // debounce after button press
      while (bdown());           // Wait for button release
      beep(100,8); delayMS(50);  // beep and debounce
      beep(100,8);
      if (k>=NPAR)               // handle exit codes
        if (k>9) { limit  = EEPROMread(k=(k-10)<<3); // recall from EEprom
                   period = EEPROMread(k+1);
                   bcd    = EEPROMread(k+2);
                   pitch  = EEPROMread(k+3);
                   tempo  = EEPROMread(k+4);
                   song   = EEPROMread(k+5);
                   break;                            // and exit from interrupt
                 }
        else     { EEPROMwrite(k=(k-8)<<3,limit = e[1] + (e[0]<<4));
                   EEPROMwrite(k+1,period = e[3] + (e[2]<<4));
                   EEPROMwrite(k+2,bcd = e[4]);      // save parameters in EEprom
                   EEPROMwrite(k+3,pitch = e[5]);
                   EEPROMwrite(k+4,tempo = e[6]);
                   EEPROMwrite(k+5,song = e[7]);
                   break;                            // and exit from interrupt
                  }
      limit = 0xE0; // upper nibble of limit will be the new parameter value
      while(1) // cycle thru the 16 possible parameter values until button push
      { PORTB = ~(k + (limit+=16)); // low/high nibble is param number/value
        for (led=200; led; led--)  // cycle at .8 seconds per value (200*.004)
        { delayMS(4); if (bdown()) goto mb; }
      }
      mb: delayMS(10);              // debounce after button press
      while (bdown());              // Wait for button release
      beep(80,15); delayMS(100); beep(120,15);
      e[k--] = limit>>4; // save new parameter value
    } // End while(1)
  }   // End if (led==SetupCode)
  pressed = 1; // indicate exit caused by button push
  led = limit; // Force exit from while loop in main
}     // End void button()
